#include <avr\io.h>
#include <avr\interrupt.h>
#define F_CPU 16000000
#include <util/delay.h>
#include <math.h>

#define DDR_COL DDRA
#define DDR_LIN DDRD
#define PORT_COL PORTA
#define PORT_LIN PORTD

#define UP    0
#define DOWN  1
#define LEFT  2
#define RIGHT 3

#define SPEED 15000

#define GIFT  0xFF
#define EMPTY 0x00

/* mesaj cu textul 'LEVEL' */
const unsigned char level_text[5][19] = 
	{1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 
	 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0,
	 1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0,
	 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0,
	 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1};

/* mesaj cu textul '1' */
const unsigned char one_text[8][8] =
	{0, 0, 0, 0, 0, 0, 0, 0,
	 0, 0, 0, 1, 1, 0, 0, 0,
	 0, 0, 0, 0, 1, 0, 0, 0,
	 0, 0, 0, 0, 1, 0, 0, 0,
	 0, 0, 0, 0, 1, 0, 0, 0,
	 0, 0, 0, 0, 1, 0, 0, 0,
	 0, 0, 0, 1, 1, 1, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0};

/* mesaj cu textul '2' */
const unsigned char two_text[8][8] = 
	{0, 0, 0, 0, 0, 0, 0, 0,
	 0, 0, 1, 1, 1, 1, 0, 0,
	 0, 0, 0, 0, 0, 1, 0, 0,
	 0, 0, 1, 1, 1, 1, 0, 0,
	 0, 0, 1, 0, 0, 0, 0, 0,
	 0, 0, 1, 0, 0, 0, 0, 0,
	 0, 0, 1, 1, 1, 1, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0};

/* mesaj cu textul '3' */
const unsigned char three_text[8][8] = 
	{0, 0, 0, 0, 0, 0, 0, 0,
	 0, 0, 1, 1, 1, 1, 0, 0,
	 0, 0, 0, 0, 0, 1, 0, 0,
	 0, 0, 1, 1, 1, 1, 0, 0,
	 0, 0, 0, 0, 0, 1, 0, 0,
	 0, 0, 0, 0, 0, 1, 0, 0,
	 0, 0, 1, 1, 1, 1, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0};

unsigned char board[8][8] = {0};	/* tabla de joc */

unsigned char enabled;				/* memoreaza daca jocul este activ sau suspendat */
unsigned char seed;					/* variabila folosita pentru generarea de numere pseudo-aleatoare */
unsigned char game_over;			/* memoreaza daca s-a terminat jocul */
unsigned char lives_left;			/* numarul de vieti ramase */
unsigned char has_gift;				/* memoreaza daca exista un cadou pe tabla de joc */
unsigned char direction;			/* directia de deplasare a sarpelui */
unsigned char gift_lin, gift_col;	/* pozitia cadoului pe tabla */
unsigned char head_lin, head_col;	/* pozitia capului sarpelui pe tabla */
unsigned char tail_lin, tail_col;	/* pozitia cozii sarpelui pe tabla */
unsigned char length;				/* lungimea sarpelui */

/**
 * Initializeaza variabilele de joc.
 */
void init_game(void)
{
	/* umple matricea cvu zerouri */
	int i, j;
	for (i = 0; i < 8; i++)
		for (j = 0; j < 8; j++)
			board[i][j] = 0;

	/* directia initiala de deplasare este spre dreapta */
	direction = RIGHT;
	
	/* lungimea initiala a sarpelui este de 4 unitati */
	length = 4;
	
	/* capul sarpelui se afla la pozitia (4, 5) */
	head_lin = 4;
	head_col = 5;

	/* coada sarpelui se afla la pozitia (4, 5) */
	tail_lin = 4;
	tail_col = 2;

	/* se initializeaza pozitiile pe care se afla sarpele */
	board[head_lin][head_col] = 1;
	board[head_lin][head_col - 1] = 2;
	board[head_lin][head_col - 2] = 3;
	board[head_lin][head_col - 3] = 4;
	
	/* cadoul se afla initial la pozitia (3, 3) */
	gift_lin = 2;
	gift_col = 3;
	board[gift_lin][gift_col] = GIFT;
	
	/* se porneste jocul */	
	seed = 91;
	has_gift = 1;
	game_over = 0;
	enabled = 1;
}

/**
 * Reseteaza tabla de joc (o umple cu zerouri).
 */
void reset(void)
{
	unsigned char i, j;
	
	for (i = 0; i < 8; i++)
		for (j = 0; j < 8; j++)
			board[i][j] = 0;
}

/**
 * Initializeaza counterul care determina viteza jocului.
 */
void init_counter(void)
{
	TCNT2 = 0;
	OCR2 = SPEED;
	TCCR2 = _BV(WGM21) | _BV(CS22) | _BV(CS21) | _BV(CS20);
	TIMSK |= _BV(OCIE2);
}

/**
 * Initializeaza butonul de pe pinul PB0 ca intrare.
 */
void init_input1(void)
{
	/* seteaza ca input pinul 0 de pe portul B si 
	 * activeaza rezistenta de pull-up */
	DDRB &= ~_BV(PB0);
	PORTB |= _BV(PB0);

	/* initializare Timer0 */
	TCCR0 = _BV(WGM01) | _BV(CS02) | _BV(CS01);
	OCR0 = 0;
	TIMSK |= _BV(OCIE0);
}

/**
 * Initializeaza butonul de pe pinul PB1 ca intrare.
 */
void init_input2(void)
{
	/* seteaza ca input pinul 1 de pe portul B si 
	 * activeaza rezistenta de pull-up */
	DDRB &= ~_BV(PB1);
	PORTB |= _BV(PB1);

	/* initializare Timer1 */
	TCCR1A = 0;
	TCCR1B |= _BV(WGM12) | _BV(CS12) | _BV(CS11);
	OCR1A = 0; 
	TIMSK |= _BV(OCIE1B); 
}

/**
 * Initializeaza porturile de iesire.
 */
void init_output(void)
{
	DDR_LIN = 0xFF;
	DDR_COL = 0xFF;
}

/**
 * Face toate initializarile necesare aplicatiei.
 */
void init(void)
{
	lives_left = 3;

	init_game();
	
	init_input1();
	init_input2();
}

/**
 * Genereaza un cadou aflat la o pozitie aleatoare pe tabla.
 */
void generate_gift(void)
{
	unsigned char i = gift_lin;
	unsigned char j = gift_col;	
	
	while (1) {
		if (i == 0) i = seed;
		if (j == 0) j = seed;
		
		i = ((unsigned short)(1000 * sin(7 * i * seed))) % 8;
		j = ((unsigned short)(1000 * cos(7 * j * seed))) % 8;
		
		if (board[i][j] == EMPTY) {
			gift_lin = i;
			gift_col = j;
			
			board[i][j] = GIFT;
			
			break;
		}
		
		seed++;
	}
	
	has_gift = 1;
}

/**
 * Incarca o coloana a tablei de joc in vederea afisarii.
 */
void load_col(unsigned char j)
{
	unsigned char i, c = 0;
	
	for (i = 0; i < 8; i++)
		if (board[j][i] != 0)
				c |= (1 << i);
		
	PORT_LIN = c;
}

/**
 * Deseneaza tabla de joc pe matricea de leduri.
 */
void draw_board(void)
{
	unsigned char i, j;
	
	while (1) {
		if (game_over) {
			for (i = 0; i < 8; i++)
				for (j = 0; j < 8; j++)
					board[i][j] = 0;
		}
	
		for (j = 0; j < 8; j++) {
			load_col(j);
			
			PORT_COL = ~(1 << j);
		}
		
		if (game_over)
			break;
	}
}

/**
 * Deseneaza textul corespunzator unui anumit nivel.
 */
void draw_level(unsigned char level)
{
	int i, j;
	static unsigned char step = 0;
	
	if (step < 30)
		for (j = 0; j < 7; j++)
			for (i = 0; i < 5; i++)
				board[i + 2][j] = board[i + 2][j + 1];
	
	if (step < 19)
		for (i = 0; i < 5; i++)
			board[i + 2][7] = level_text[i][step];
	else 
	if (step < 30)
		for (i = 0; i < 5; i++)
			board[i + 2][7] = 0;
	else 
	if (step == 31) {
		for (i = 0; i < 8; i++)
			for (j = 0; j < 8; j++) {
				if (level == 1) (board[i][j] = one_text[i][j]); else
				if (level == 2) (board[i][j] = two_text[i][j]); else
				if (level == 3) (board[i][j] = three_text[i][j]);
			}
	}
	else
	if (step == 36) {
		for (i = 0; i < 8; i++)
			for (j = 0; j < 8; j++)
				board[i][j] = 0;
	}
	if (step > 35)
		init();
	
	step++;
}

/**
 * Intoarce 1 daca la urmatorul pas se va produce o coliziune
 * intre capul sarpelui si o parte a corpului. Intoarce 0 in
 * caz contrar.
 */
unsigned char detect_collision(void)
{
	switch (direction) {
		case UP: 
			if (head_lin > 0) {
				if (board[head_lin - 1][head_col] >= 1 && board[head_lin - 1][head_col] <= length)
					return 1;
			}
			else
				if (board[7][head_col] >= 1 && board[7][head_col] <= length)
					return 1;
		break;
		case DOWN:
			if (head_lin < 7) {
				if (board[head_lin + 1][head_col] >= 1 && board[head_lin + 1][head_col] <= length)
					return 1;
			}
			else
			if (board[0][head_col] >= 1 && board[0][head_col] <= length)
				return 1;
		break;
		case LEFT:
			if (head_col > 0) {
				if (board[head_lin][head_col - 1] >= 1 && board[head_lin][head_col - 1] <= length)
					return 1;
			}
			else
			if (board[head_lin][7] >= 1 && board[head_lin][7] <= length)
				return 1;
		break;
		case RIGHT:
			if (head_col < 7) {
				if (board[head_lin][head_col + 1] >= 1 && board[head_lin][head_col + 1] <= length)
					return 1;
			}
			else
			if (board[head_lin][0] >= 1 && board[head_lin][0] <= length)
				return 1;
		break;
	}
	
	return 0;
}

/**
 * Misca sarpele pe tabla.
 */
void move_snake(void)
{
	unsigned char i, j;
	unsigned char tail_lin_aux, tail_col_aux;

	/* daca are loc o colizune jocul este resetat */
	if (detect_collision()) {
		reset();
		lives_left--;
		
		if (lives_left == 0)
			game_over = 1;
		else
			init_game();
		
		return;
	}
	
	/* scaneaza tabla de joc pentru gasirea penultimului element al sarpelui */	
	for (i = 0; i < 8; i++)
		for (j = 0; j < 8; j++)
			if (board[i][j] == length - 1) {
				tail_lin_aux = tail_lin;
				tail_col_aux = tail_col;
				
				break;
			}
		
	/* muta fiecare element al corpului sarpelui */
	for (i = 0; i < 8; i++)
		for (j = 0; j < 8; j++)
			if (board[i][j] != EMPTY && board[i][j] != GIFT)
				if (board[i][j] != length)					
					board[i][j]++;
				else
					board[i][j] = EMPTY;
	
	tail_lin = tail_lin_aux;
	tail_col = tail_col_aux;
	
	/* muta capul sarpelui */
	switch (direction) {
		case UP: 
			if (head_lin > 0)
				head_lin--;
			else
				head_lin = 7;
		break;
		case DOWN: 
			if (head_lin < 7)
				head_lin++;
			else
				head_lin = 0;
		break;
		case LEFT:
			if (head_col > 0)
				head_col--;
			else
				head_col = 7;
		break;
		case RIGHT:
			if (head_col < 7)
				head_col++;
			else
				head_col = 0;
		break;
	}

	board[head_lin][head_col] = 1;

	/* testeaza daca a fost colectat un cadou */
	if (head_lin == gift_lin && head_col == gift_col) {		
		for (i = 0; i < 8; i++)
			for (j = 0; j < 8; j++)
				if (board[i][j] != EMPTY && board[i][j] != GIFT)
					board[i][j]++;
		
		switch (direction) {
			case UP: 
				if (head_lin > 0)
					head_lin--;
				else
					head_lin = 7;
			break;
			case DOWN: 
				if (head_lin < 7)
					head_lin++;
				else
					head_lin = 0;
			break;
			case LEFT:
				if (head_col > 0)
					head_col--;
				else
					head_col = 7;
			break;
			case RIGHT:
				if (head_col < 7)
					head_col++;
				else
					head_col = 0;
			break;
		}
		
		board[head_lin][head_col] = 1;
		
		length++;
		
		generate_gift();
	}
}

/**
 * Rutina de apasarea a butonului de pe PB0.
 */
ISR(TIMER0_COMP_vect)
{
	if (enabled) {
		if (direction == UP || direction == DOWN)
			direction = RIGHT;
		else
		if (direction == LEFT || direction == RIGHT)
			direction = UP;
	}
}

/**
 * Rutina de apasarea a butonului de pe PB1.
 */
ISR(TIMER1_COMPB_vect)
{
	if (enabled) {
		if (direction == UP || direction == DOWN)
			direction = LEFT;
		else
		if (direction == LEFT || direction == RIGHT)
			direction = DOWN;		
	}
}

/**
 * Rutina counterului (deplasarea sarpelui la un interval de timp).
 */
ISR(TIMER2_COMP_vect)
{
	if (enabled)
		move_snake();
	else
		draw_level(1);
}

int main()
{
	sei();
	init_output();
	init_counter();
	
	enabled = 0;
	
    draw_board();
	
	return 0;
}
